
package jp.gr.java_conf.u6k.andro_memo.model;

import java.util.List;
import java.util.Properties;

import javax.jdo.PersistenceManager;
import javax.jdo.Query;
import javax.mail.Message;
import javax.mail.MessagingException;
import javax.mail.Session;
import javax.mail.Transport;
import javax.mail.internet.AddressException;
import javax.mail.internet.InternetAddress;
import javax.mail.internet.MimeMessage;

import jp.gr.java_conf.u6k.andro_memo.entity.LoginTicketEntity;
import jp.gr.java_conf.u6k.andro_memo.entity.MemoEntity;
import jp.gr.java_conf.u6k.andro_memo.entity.UserEntity;
import jp.gr.java_conf.u6k.andro_memo.entity.UserStateEnum;
import jp.gr.java_conf.u6k.andro_memo.util.CollectionUtil;
import jp.gr.java_conf.u6k.andro_memo.util.DateUtil;
import jp.gr.java_conf.u6k.andro_memo.util.HashUtil;

import com.google.appengine.api.datastore.KeyFactory;

public class UserModel {

    // private static final Logger LOG = Logger.getLogger(UserModel.class.getName());

    private PersistenceManager _pm;

    public UserModel(PersistenceManager pm) {
        this._pm = pm;
    }

    /**
     * 会員仮登録を行います。
     * <ul>
     * <li>ユーザーを仮登録状態でデータ作成します。
     * <li>会員仮登録メールを送信します。
     * </ul>
     * 
     * @param email
     *            仮登録を行うユーザーのメールアドレス。
     */
    public void registTemp(String email) {
        // 入力チェックをします。メールアドレスがメールアドレスの形式か調べます(.+@.+)。
        if (email == null) {
            throw new IllegalArgumentException("email == null");
        }
        if (!email.matches("^.+@.+$")) {
            throw new IllegalArgumentException("email.format illegal");
        }

        // 同一メールアドレスを持つユーザーが存在するか調べます。
        Query query = this._pm.newQuery(UserEntity.class);
        query.setFilter("email == emailParam");
        query.declareParameters("String emailParam");

        List<UserEntity> userList = CollectionUtil.castToList(query.execute(email));

        if (userList.size() > 0) {
            throw new UserAlreadyExistedException();
        }

        // ユーザーを仮登録状態でデータ作成します。
        UserEntity user = new UserEntity();
        user.setEmail(email);
        user.setUpdateDatetime(DateUtil.getNow(DateUtil.getUserTimeZone()));
        user.setState(UserStateEnum.TEMPORARY);
        this._pm.makePersistent(user);

        // 会員仮登録メールを送信します。
        String from = "u6k.yu1@gmail.com"; // TODO 自動送信メールアドレスに変更する。
        String subject = "[AndroMemo]会員登録受け付けのお知らせ";
        String body = "会員を仮登録してくださいまして、ありがとうございます。\r\n";
        body += "引き続き会員登録を行うため、以下のURLにアクセスしてください。\r\n";
        body += "\r\n";
        body += "http://andro-memo.appspot.com/regist.servlet?key=" + KeyFactory.keyToString(user.getKey()) + "\r\n";

        Properties props = System.getProperties();
        Session mailSession = Session.getDefaultInstance(props, null);

        try {
            InternetAddress mailFrom = new InternetAddress(from);
            InternetAddress mailTo = new InternetAddress(email);

            MimeMessage mail = new MimeMessage(mailSession);
            mail.setHeader("Content-Type", "text/plain; charset=ISO-2022-JP");
            mail.setHeader("Content-Transfer-Encoding", "7bit");
            mail.setFrom(mailFrom);
            mail.addRecipient(Message.RecipientType.TO, mailTo);
            mail.setSubject(subject, "ISO-2022-JP");
            mail.setText(body, "ISO-2022-JP");

            Transport.send(mail);
        } catch (AddressException e) {
            throw new SendMailException(e);
        } catch (MessagingException e) {
            throw new SendMailException(e);
        }
    }

    /**
     * 仮登録状態のユーザーを取得します。
     * <ul>
     * <li>キーを使用して、仮登録状態のユーザーを検索します。
     * </ul>
     * 
     * @param key
     *            取得する仮登録ユーザーのキー。
     * @return 仮登録状態のユーザー。
     */
    public UserEntity findTempUser(String key) {
        UserEntity user;

        // キーを使用して、仮登録状態のユーザーを検索します。
        user = this.findUserByKey(key);

        if (user.getState() != UserStateEnum.TEMPORARY) {
            throw new UserNotFoundException("user.state != TEMPORARY.");
        }

        return user;
    }

    /**
     * 仮登録状態のユーザーを正式登録状態に更新します。
     * <ul>
     * <li>仮登録ユーザーのプロパティを更新します。
     * <li>ユーザーを更新する。
     * <li>会員登録完了メールを送信します。
     * </ul>
     * 
     * @param tempUser
     *            仮登録ユーザー。
     * @param password
     *            ユーザーに設定するパスワード。
     */
    public void regist(String key, String password) {
        // 入力チェックをします。
        if (password == null) {
            throw new IllegalArgumentException("password == null");
        }
        if (!password.matches("^[0-9a-zA-Z]{7,}$")) {
            throw new IllegalArgumentException("password.format illegal");
        }

        // 仮登録ユーザーのプロパティを更新します。
        // 　　パスワード：設定
        // 　　状態：仮登録→活性
        UserEntity user = this.findTempUser(key);

        String passwordHash = HashUtil.digest(user.getPasswordHashAlgorithm(), password);

        user.setPasswordHash(passwordHash);
        user.setState(UserStateEnum.LIVES);
        user.setUpdateDatetime(DateUtil.getNow(DateUtil.getUserTimeZone()));

        this._pm.makePersistent(user);

        // 会員登録完了メールを送信します。
        String from = "u6k.yu1@gmail.com"; // TODO 自動送信メールアドレスに変更する。
        String subject = "[AndroMemo]会員登録完了のお知らせ";
        String body = "会員登録が完了しました。\r\n";
        body += "AndroMemoで快適なメモ生活をお送りください。\r\n";

        Properties props = System.getProperties();
        Session mailSession = Session.getDefaultInstance(props, null);

        try {
            InternetAddress mailFrom = new InternetAddress(from);
            InternetAddress mailTo = new InternetAddress(user.getEmail());

            MimeMessage mail = new MimeMessage(mailSession);
            mail.setHeader("Content-Type", "text/plain; charset=ISO-2022-JP");
            mail.setHeader("Content-Transfer-Encoding", "7bit");
            mail.setFrom(mailFrom);
            mail.addRecipient(Message.RecipientType.TO, mailTo);
            mail.setSubject(subject, "ISO-2022-JP");
            mail.setText(body, "ISO-2022-JP");

            Transport.send(mail);
        } catch (AddressException e) {
            throw new SendMailException(e);
        } catch (MessagingException e) {
            throw new SendMailException(e);
        }
    }

    /**
     * 退会します。
     * <ul>
     * <li>ログイン・チケットの持ち主であるユーザーを検索します。
     * <li>ユーザー、及びユーザーが持つデータを全て削除します。
     * <li>退会メールを送信します。
     * </ul>
     * 
     * @param loginTicketKey
     *            退会するユーザーのログイン・チケット・キー。
     */
    public void unregist(String loginTicketKey) {
        // 入力チェックをします。
        if (!this.isLogin(loginTicketKey)) {
            throw new UserNotLoginedException();
        }

        String email;

        // ログイン・チケットの持ち主であるユーザーを検索します。
        String userKey = KeyFactory.keyToString(KeyFactory.stringToKey(loginTicketKey).getParent());

        UserEntity user = this.findUserByKey(userKey);

        email = user.getEmail();

        // ユーザーが持つメモを検索する。
        Query query = this._pm.newQuery(MemoEntity.class);
        query.setFilter("user == userParam");
        query.declareParameters("UserEntity userParam");

        List<MemoEntity> memoList = CollectionUtil.castToList(query.executeWithArray(user));

        // ユーザーが持つログイン・チケットを検索する。
        query = this._pm.newQuery(LoginTicketEntity.class);
        query.setFilter("user == userParam");
        query.declareParameters("UserEntity userParam");

        List<LoginTicketEntity> loginTicketList = CollectionUtil.castToList(query.executeWithArray(user));

        // ユーザー、及びユーザーが持つすべてのデータを削除します。
        this._pm.deletePersistent(user);

        this._pm.deletePersistentAll(memoList);

        this._pm.deletePersistentAll(loginTicketList);

        // 退会メールを送信します。
        String from = "u6k.yu1@gmail.com"; // TODO 自動送信メールアドレスに変更する。
        String subject = "[AndroMemo]退会完了のお知らせ";
        String body = "退会が完了しました。\r\n";
        body += "AndroMemoのまたのご利用をお待ちしております。\r\n";

        Properties props = System.getProperties();
        Session mailSession = Session.getDefaultInstance(props, null);

        try {
            InternetAddress mailFrom = new InternetAddress(from);
            InternetAddress mailTo = new InternetAddress(email);

            MimeMessage mail = new MimeMessage(mailSession);
            mail.setHeader("Content-Type", "text/plain; charset=ISO-2022-JP");
            mail.setHeader("Content-Transfer-Encoding", "7bit");
            mail.setFrom(mailFrom);
            mail.addRecipient(Message.RecipientType.TO, mailTo);
            mail.setSubject(subject, "ISO-2022-JP");
            mail.setText(body, "ISO-2022-JP");

            Transport.send(mail);
        } catch (AddressException e) {
            throw new SendMailException(e);
        } catch (MessagingException e) {
            throw new SendMailException(e);
        }
    }

    /**
     * ログインします。
     * 
     * @param email
     *            ログインするユーザーのメールアドレス。
     * @param password
     *            ログインするユーザーのパスワード。
     * @return ログイン・チケット。
     */
    public String login(String email, String password) {
        // 入力チェックをします。
        if (email == null) {
            throw new IllegalArgumentException("email == null");
        }
        if (password == null) {
            throw new IllegalArgumentException("password == null");
        }

        LoginTicketEntity loginTicket;

        // メールアドレス、パスワードを使用して、活性状態のユーザーを検索します。
        Query query = this._pm.newQuery(UserEntity.class);
        query.setFilter("email == emailParam && state == stateParam");
        query.declareParameters("String emailParam, UserStateEnum stateParam");

        List<UserEntity> userList = CollectionUtil.castToList(query.executeWithArray(email, UserStateEnum.LIVES));
        if (userList.size() == 0) {
            throw new UserNotFoundException("email != emailParam || state != LIVES");
        }

        UserEntity user = userList.get(0);

        String passwordHash = HashUtil.digest(user.getPasswordHashAlgorithm(), password);

        if (!user.getPasswordHash().equals(passwordHash)) {
            throw new UserNotFoundException("password != passwordParam");
        }

        // 検索したユーザーを使用して、ログイン・チケットをデータ作成します。
        loginTicket = new LoginTicketEntity();
        loginTicket.setLoginDatetime(DateUtil.getNow(DateUtil.getUserTimeZone()));
        loginTicket.setLastAccessDatetime(DateUtil.getNow(DateUtil.getUserTimeZone()));

        user.getLoginTicketList().add(loginTicket);

        this._pm.makePersistent(user);

        String loginTicketId = KeyFactory.keyToString(loginTicket.getKey());

        return loginTicketId;
    }

    /**
     * ログイン・チケットがログイン状態かどうかを調べます。また、ログイン期限が切れている場合、ログアウトします。
     * 
     * @param loginTicketId
     *            ログイン状態かどうかを調べるログイン・チケット。
     * @return ログイン状態の場合はtrue。ログアウト状態の場合、またはログイン期限が切れた場合、false。
     */
    public boolean isLogin(String loginTicketKey) {
        // 1.ログイン・チケットIDに一致する、ログイン状態のログイン・チケットを検索する。
        Query query = this._pm.newQuery(LoginTicketEntity.class);
        query.setFilter("key == keyParam");
        query.declareParameters("String keyParam");

        List<LoginTicketEntity> loginTicketList = CollectionUtil.castToList(query.executeWithArray(loginTicketKey));

        // 1-1.ログイン・チケットが存在しない場合、LoginTicketNotFoundException例外をスローする。
        if (loginTicketList.size() == 0) {
            throw new LoginTicketNotFoundException();
        }

        LoginTicketEntity loginTicket = loginTicketList.get(0);

        // 2.ログイン期限が超過していないか調べる。
        long loginTime = loginTicket.getLoginDatetime().getTime();
        long timeoutTime = loginTime + (1000 * 60 * 60 * 24 * 7);
        long nowTime = DateUtil.getNow(DateUtil.getUserTimeZone()).getTime();

        boolean timeout = (timeoutTime < nowTime);

        // 2-1.ログイン期限が超過している場合、ログイン・チケットを削除します。
        if (timeout) {
            this._pm.deletePersistent(loginTicket);

            this._pm.makePersistent(loginTicket);
        }

        return true;
    }

    /**
     * ログイン・チケットを使用して、ログインしているユーザーを取得します。
     * 
     * @param loginTicketKey
     *            ログイン・チケットのキー。
     * @return ログインしているユーザー。
     */
    public UserEntity findLoginUser(String loginTicketKey) {
        if (!this.isLogin(loginTicketKey)) {
            throw new UserNotLoginedException();
        }

        // ログイン・チケットの親であるユーザーを取得します。
        String userKey = KeyFactory.keyToString(KeyFactory.stringToKey(loginTicketKey).getParent());

        UserEntity user = this.findUserByKey(userKey);

        return user;
    }

    /**
     * ログアウトします。
     * 
     * @param loginTicketKey
     *            ログイン・チケットのキー。
     */
    public void logout(String loginTicketKey) {
        // ログイン状態のチェックをします。
        if (!this.isLogin(loginTicketKey)) {
            throw new UserNotLoginedException();
        }

        // ログイン・チケットを検索します。
        Query query = this._pm.newQuery(LoginTicketEntity.class);
        query.setFilter("key == keyParam");
        query.declareParameters("String keyParam");

        List<LoginTicketEntity> loginTicketList = CollectionUtil.castToList(query.executeWithArray(loginTicketKey));

        if (loginTicketList.size() == 0) {
            throw new LoginTicketNotFoundException();
        }

        LoginTicketEntity loginTicket = loginTicketList.get(0);

        // ログイン・チケットを削除します。
        this._pm.deletePersistent(loginTicket);
    }

    private UserEntity findUserByKey(String key) {
        UserEntity user;

        Query query = this._pm.newQuery(UserEntity.class);
        query.setFilter("key == keyParam");
        query.declareParameters("String keyParam");

        List<UserEntity> userList = CollectionUtil.castToList(query.execute(key));

        if (userList.size() == 0) {
            throw new UserNotFoundException("key != keyParam");
        } else {
            user = userList.get(0);
        }

        return user;
    }

}
